//  ===================================================================
//
//  Program:   2D Mesh Viewer (MeshView)
//  Module:    control.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "control.h"
#include <sys/time.h>

namespace meshview {

  // -----------------------------------------------------------------
  // Method drawMesh()
  //
  // Draws a  2D mesh in an  OpenGL window. Constrained  edges will be
  // displayed with thicker lines.
  //
  // -----------------------------------------------------------------
  void
  tControl::drawMesh()
  {
    if (m_mesh == 0) return;

    cpPointer cp = m_mesh->getCellComplex();

    // Draw edges
    egPointer e;
    hePointer he;
    vtPointer v1;
    vtPointer v2;

    for (EIT ei = cp->egBegin(); ei != cp->egEnd(); ++ei) {
      e = *ei;
      he = e->getHalfEdge();
      v1 = he->getVertex();
      v2 = he->getNext()->getVertex();

      double x1 = v1->getPoint().x();
      double y1 = v1->getPoint().y();
      double x2 = v2->getPoint().x();
      double y2 = v2->getPoint().y();

      if (e->isConstrained()) {
	glLineWidth(2.0);
	glBegin(GL_LINES);
	glColor3f(0.0,1.0,0.0);
	glVertex2f(x1,y1);
	glVertex2f(x2,y2);
	glEnd();
	glLineWidth(1.0);
      }
      else {
	glBegin(GL_LINES);
	glColor3f(0.7,0.7,0.7);
	glVertex2f(x1,y1);
	glVertex2f(x2,y2);
	glEnd();
      }

      glBegin(GL_POINTS);
      glColor3f(1.0,0.0,0.0);
      glVertex2f(x1,y1);
      glVertex2f(x2,y2);
      glEnd();
    }
  }

  // -----------------------------------------------------------------
  // Method buildQuadMesh()
  //
  // Builds a quadrilateral mesh from files .NODE, .EDGE and .ELE.
  //
  // -----------------------------------------------------------------
  void
  tControl::buildQuadMesh() {
    if (m_mesh == 0) {
      m_mesh = new tQuadMesh();

      try {
	m_mesh->buildMesh(m_fname);
      }
      catch (const tExceptionObject& xpt) {
	treatException(xpt);
	exit(0);
      }
      
      cpPointer cp = m_mesh->getCellComplex();

      std::cout << "Quadrangulation has successfully been built." << std::endl;

      std::cout << std::endl;

      std::cout << "Number of vertices  : " << cp->getNumVerts()  << std::endl;
      std::cout << "Number of edges     : " << cp->getNumEdges()  << std::endl;
      std::cout << "Number of faces     : " << cp->getNumFaces()  << std::endl;
      std::cout << "Number of holes     : " << cp->getNumHoles()  << std::endl;
      std::cout << "Number of boundaries: " << cp->getNumBounds() << std::endl;

      std::cout << std::endl;
    }
  }

  // -----------------------------------------------------------------
  // Method saveToEPSFile()
  //
  // Generates an  EPS file  corresponding to the  mesh showed  on the
  // screen.
  //
  // -----------------------------------------------------------------
  void
  tControl::saveToEPSFile() {

    // If there is no mesh, leave the method
    if (m_mesh == 0) return;

    // Gets a pointer to the mesh
    cpPointer cp = m_mesh->getCellComplex();

    // Gets the bounding box of the mesh
    tBBox2::Pointer bb(this->getMeshBoundingBox());
    Point2 pmin = bb->getPMin();
    Point2 pmax = bb->getPMax();

    // Gets the bounding box coordinates
    double xmin = pmin.x();
    double xmax = pmax.x();
    double ymin = pmin.y();
    double ymax = pmax.y();

    // Computes the EPS file bounding box
    double dx = xmax - xmin;
    double dy = ymax - ymin;
    int llx, urx, lly, ury;
    const double XDIM = 612;
    const double YDIM = 792;
    const double LENGTH = 512;
    if (dx >= dy) {
      llx = (int) ((XDIM - LENGTH) / 2);
      urx = (int) (llx + LENGTH + 1);
      lly = (int) (((YDIM - ((dy / dx) * LENGTH))) / 2);
      ury = (int) (lly + ((dy / dx) * LENGTH) + 1);
    }
    else {
      llx = (int) ((XDIM - ((dx / dy) * LENGTH)) / 2);
      urx = (int) (llx + ((dx / dy) * LENGTH) + 1);
      lly = (int) ((YDIM - LENGTH) / 2);
      ury = (int) (lly + LENGTH + 1);
    }


    // Creates an output file
    std::string fname = m_fname + std::string(".eps");

    std::filebuf fb;
    if (fb.open(fname.c_str(), std::ios::out) == 0) {
      std::stringstream ss (std::stringstream::in | 
			    std::stringstream::out);
      ss << "Unable to create output file " << fname.c_str();
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    std::ostream ostr(&fb);

    //
    // Writes an EPS file header
    //
    ostr << "%%!PS-Adobe-2.0 EPSF-2.0" << std::endl;
    ostr << "%%BoundingBox: " << llx << " " << lly << " " 
	 << urx << " " << ury << std::endl;
    ostr << "%%Creator: MeshView" << std::endl;
    ostr << "%%EndComments" << std::endl << std::endl;
    ostr << "1 setlinecap" << std::endl;
    ostr << "1 setlinejoin" << std::endl;
    ostr << "1 setlinewidth" << std::endl;
    ostr << llx << " " << lly << " moveto" << std::endl;
    ostr << urx << " " << lly << " lineto" << std::endl;
    ostr << urx << " " << ury << " lineto" << std::endl;
    ostr << llx << " " << ury << " lineto" << std::endl;
    ostr << "closepath" << std::endl
	 << "clip"      << std::endl
	 << "newpath"   << std::endl;

    //
    // Writes the edges of the mesh
    //

    egPointer e;
    hePointer he;
    vtPointer v1;
    vtPointer v2;
    for (EIT ei = cp->egBegin(); ei != cp->egEnd(); ++ei) {
      e = *ei;
      he = e->getHalfEdge();
      v1 = he->getVertex();
      v2 = he->getNext()->getVertex();

      double x1 = v1->getPoint().x();
      double y1 = v1->getPoint().y();
      double x2 = v2->getPoint().x();
      double y2 = v2->getPoint().y();

      x1 = ((x1 - xmin) * (urx - llx)) / (xmax - xmin) + llx;
      y1 = ((y1 - ymin) * (ury - lly)) / (ymax - ymin) + lly;
      x2 = ((x2 - xmin) * (urx - llx)) / (xmax - xmin) + llx;
      y2 = ((y2 - ymin) * (ury - lly)) / (ymax - ymin) + lly;

      if (e->isConstrained()) {
	ostr << "2 setlinewidth" << std::endl;
	ostr << x1 << " " << y1 << " moveto" << std::endl;
	ostr << x2 << " " << y2 << " lineto" << std::endl;
	ostr << "stroke"  << std::endl;
      }
      else {
	ostr << "1 setlinewidth" << std::endl;
	ostr << x1 << " " << y1 << " moveto" << std::endl;
	ostr << x2 << " " << y2 << " lineto" << std::endl;
	ostr << "stroke"  << std::endl; 
      }
    }

    //
    // Writes the vertices of the mesh
    //

    vtPointer v;
    for (VIT vi = cp->vtBegin(); vi != cp->vtEnd(); ++vi) {
      v = *vi;
      double x = v->getPoint().x();
      double y = v->getPoint().y();

      x = ((x - xmin) * (urx - llx)) / (xmax - xmin) + llx;
      y = ((y - ymin) * (ury - lly)) / (ymax - ymin) + lly;

      ostr << x << " " << y << " 1.3 0 360 arc" << std::endl;
      ostr <<   "fill" << std::endl;
      ostr << "stroke" << std::endl;
    }

    fb.close();
  }

}   // end namespace meshview
