//  ===================================================================
//
//  Program:   2D Mesh Viewer (MeshView)
//  Module:    quadmesh2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
// ===================================================================

#include "quadmesh2.h"

namespace meshview {

  // -----------------------------------------------------------------
  // Method buildQuadMesh()
  //
  // Stores  a  quadrilateral  mesh  into the  data  structure  handle
  // edge. All  it needs  is the collection  of quadrilaterals  of the
  // mesh in  any given  order. Quadrilaterals should  consistently be
  // oriented.
  //
  // -----------------------------------------------------------------
  tMesh::cpPointer
  tQuadMesh::buildMesh(const std::string& name) 
    throw (tExceptionObject)
  {
    // Reads  files   with  information  about   vertices,  edges  and
    // triangles of the triangular mesh.
    std::string infile1 = std::string(name) + std::string(".node");
    std::string infile2 = std::string(name) + std::string(".edge");
    std::string infile3 = std::string(name) + std::string(".ele");

    // Creates file reader
    tMeshReader reader(infile1, infile2, infile3);

    // Reads files
    NodeSet lv;
    ConstrainedEdgeSet le;
    FaceSet lf;

    try {
      reader.readVertexFile(lv);
      reader.readEdgeFile(le);
      reader.readElemFile(lv,lf);
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    // We do not need the set  of nodes anymore as node information is
    // also in the face set "lf".
    lv.clear();

    // First stage: making faces
    GlueEdgeSet gluelist;
    typedef FaceSet::const_iterator LFI;

    for (LFI fiter = lf.begin(); fiter != lf.end(); ++fiter) {
      // Gets face from iterator
      tQuadrilateral2 f(*fiter);

      // Checks out if face is simple
      if (!f.isSimple()) {
	std::stringstream ss (std::stringstream::in | std::stringstream::out);
	ss << "Quadrilateral is not simple";
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Checks out orientation
      if (!f.isCCW()) {
	f.reverse();
      }

      // Creates vertex codes for every vertex of face f
      int cv1 = f.getVertexId(0);
      int cv2 = f.getVertexId(1);
      int cv3 = f.getVertexId(2);
      int cv4 = f.getVertexId(3);

      // Gets vertex locations
      Point2 lv1 = f.getPoint(0);
      Point2 lv2 = f.getPoint(1);
      Point2 lv3 = f.getPoint(2);
      Point2 lv4 = f.getPoint(3);

      // Creates edge codes for every edge of face f
      tEdgeKey ce1 = tEdgeKey(cv1,cv2);
      tEdgeKey ce2 = tEdgeKey(cv2,cv3);
      tEdgeKey ce3 = tEdgeKey(cv3,cv4);
      tEdgeKey ce4 = tEdgeKey(cv4,cv1);

      // Searches for existing edges in the boundary edge set
      int foundedge = 0;
      if (es.find(ce1) != es.end()) foundedge += 1;
      if (es.find(ce2) != es.end()) foundedge += 2;
      if (es.find(ce3) != es.end()) foundedge += 4;
      if (es.find(ce4) != es.end()) foundedge += 8;

      // A new  face is created only if  one of its edges  is found in
      // the set of constrained edges  and if the attaching of the new
      // face does not create a singular vertex.
      if ((foundedge == 1) || (foundedge == 2) || (foundedge == 4) || 
	  (foundedge == 8)) {
	int cvaux1, cvaux2;
	Point2 lvaux1, lvaux2;
	tEdgeKey ceaux1, ceaux2, ceaux3, ceaux4;

	if (foundedge == 1) {
	  cvaux1 = cv3;
	  lvaux1 = lv3;
	  cvaux2 = cv4;
	  lvaux2 = lv4;

	  ceaux1 = ce2;
	  ceaux2 = ce3; 
	  ceaux3 = ce4;
	  ceaux4 = ce1;
	}
	else if (foundedge == 2) {
	  cvaux1 = cv4;
	  lvaux1 = lv4;
	  cvaux2 = cv1;
	  lvaux2 = lv1;

	  ceaux1 = ce3;
	  ceaux2 = ce4;
	  ceaux3 = ce1;
	  ceaux4 = ce2;
	}
	else if (foundedge == 4) {
	  cvaux1 = cv1;
	  lvaux1 = lv1;
	  cvaux2 = cv2;
	  lvaux2 = lv2;

	  ceaux1 = ce4;
	  ceaux2 = ce1;
	  ceaux3 = ce2;
	  ceaux4 = ce3;
	}
	else {  // if (foundedge == 8)
	  cvaux1 = cv2;
	  lvaux1 = lv2;
	  cvaux2 = cv3;
	  lvaux2 = lv3;

	  ceaux1 = ce1;
	  ceaux2 = ce2;
	  ceaux3 = ce3;
	  ceaux4 = ce4;
	}

	// If  vertices "cvaux1"  and  "cvaux2" are  not  in the  same
	// boundary as edge "ceaux4.reverse()", we can attach face "f"
	// to the boundary containing "ceaux4.reverse()".
	if ((!findVertexInBound(cvaux1,ceaux4)) && 
	    (!findVertexInBound(cvaux2,ceaux4))) {

	  // Either "cvaux1" and "cvaux2" have not been created yet or
	  // they are not on the same boundary as "ceaux4.reverse()".
	  this->makeQuadFace(f,cvaux1,lvaux1,cvaux2,lvaux2,ceaux4,
	    ceaux1,ceaux2,ceaux3,le);
	}
	else {
	  // Creates an isolated face, which should be attached to "f"
	  // in the second stage.

	  this->makeQuadFaceGlue1(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	    foundedge,gluelist,le);
	}
      }
      else if ((foundedge ==  3) || (foundedge ==  5) || 
	       (foundedge ==  9) || (foundedge ==  6) ||
	       (foundedge == 10) || (foundedge == 12)) {
	// Face  "f" has  two edges  in common  with the  current cell
	// complex. So, we make "f" an isolated face, and then we mark
	// the common edges as edges to be glued later.
	this->makeQuadFaceGlue2(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	  ce1,ce2,ce3,ce4,foundedge,gluelist,le);
      }
      else if ((foundedge ==  7) || (foundedge == 11) || 
	       (foundedge == 13) || (foundedge == 14)) {
	// New face  has three edges  in common with the  current cell
	// complex. So, we make "f" an isolated face, and then we mark
	// the common edges as edges to be glued later.
	this->makeQuadFaceGlue3(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	  ce1,ce2,ce3,ce4,foundedge,gluelist,le);
      }
      else if (foundedge != 0) {
	// New face  has four  edges in common  with the  current cell
	// complex. So, we make "f" an isolated face, and then we mark
	// the common edges as edges to be glued later.
	this->makeQuadFaceGlue4(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	  ce1,ce2,ce3,ce4,gluelist,le);
      }
      else { // foundedge == 0
	// New face  has no  vertex in common  with existent  faces or
	// one, two,  three or four  vertices in common  with existent
	// faces. We will make an isolated faces in any case.

	this->makeQuadFace(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,ce1,ce2,
			   ce3,ce4,le);
      }
    }

    // Removes all elements from vertex and edge boundary sets
    vs.clear();
    es.clear();

    // Removes all elements from constrained edge set

    // Second stage: glueing faces
    typedef GlueEdgeSet::iterator LGI;
    LGI giter = gluelist.begin();
    while (giter != gluelist.end()) {
      // Gets the half-edges to be identified in the faces to be glued.
      hePointer he1 = (*giter).m_he1;
      hePointer he2 = (*giter).m_he2;

      // Glues faces
      glueFaces(he1,he2);

      // Removes the half-edges from the glue list
      giter = gluelist.erase(giter);
    }

    // Verifies mesh consistency
    std::cout << "Checking mesh consistency..." << std::endl;
    if (!checkMeshConsistency()) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Quadrilateral mesh is not consistent";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    std::cout << "Quadrilateral mesh is consistent." << std::endl;

    // Returns complex
    return m_cp;
  }

  // -----------------------------------------------------------------
  // Method makeQuadFace()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFace(
			const tQuadrilateral2& f,
			int cv1,
			int cv2,
			int cv3,
			int cv4,
			const Point2& p1,
			const Point2& p2,
			const Point2& p3,
			const Point2& p4,
			const tEdgeKey& ce1,
			const tEdgeKey& ce2,
			const tEdgeKey& ce3,
			const tEdgeKey& ce4,
			const ConstrainedEdgeSet& ledges
			)
  {
    // Creates all three vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
 
    // Inserts the code of the new vertices into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce1.reverse(),he1));
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));
    es.insert(make_pair(ce4.reverse(),he4));

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeQuadFace()
  //
  // Creates a new triangular face  and attaches it to another face in
  // the mesh.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFace(
			const tQuadrilateral2& f,
			int cv1,
			const Point2& p1,
			int cv2,
			const Point2& p2,
			const tEdgeKey& ce,
			const tEdgeKey& ce2,
			const tEdgeKey& ce3,
			const tEdgeKey& ce4,
			const ConstrainedEdgeSet& le
			)
  {
    // Creates two new vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);

    // Gets first half-edge of the commom edge
    hePointer he = (es.find(ce))->second;

    // Makes a new face with a commom edge and a new edge
    egPointer e = ((m_cp->mkFE(he))->getHalfEdge())->getEdge();

    // Gets first half-edge of the new edge
    hePointer he2 = e->getHalfEdge();

    // Creates  edge following  the  common edge  and  gets its  first
    // half-edge.
    hePointer he3 = (m_cp->mkEV(v1,he2))->getHalfEdge();

    // Creates  edge preceding  the  common edge  and  gets its  first
    // half-edge.
    hePointer he4 = (m_cp->mkEV(v2,he3))->getHalfEdge();

    // Inserts the code of the new vertices into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));
    es.insert(make_pair(ce4.reverse(),he4));

    // Removes commom edge from the boundary edge set
    es.erase(es.find(ce));

    // Sets status of constrained edges
    if (isConstrainedEdge(ce2,le)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,le)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,le)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeTriFaceGlue1()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh and keeps track of  its common edge with another face of
  // the mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh:: makeQuadFaceGlue1(
				const tQuadrilateral2& f,
				int cv1,
				int cv2,
				int cv3,
				int cv4,
				const Point2& p1,
				const Point2& p2,
				const Point2& p3,
				const Point2& p4,
				int cedge,
				GlueEdgeSet& gset,
				const ConstrainedEdgeSet& ledges
			      )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);
    
    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    EdgeSet::iterator hepos;
    tEdgeKey ce1(cv2,cv1);
    tEdgeKey ce2(cv3,cv2);
    tEdgeKey ce3(cv4,cv3);
    tEdgeKey ce4(cv1,cv4);
    if (cedge == 1) {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce2,he2));
      es.insert(make_pair(ce3,he3));
      es.insert(make_pair(ce4,he4));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce1.reverse());
      insertGlue(hepos->second,he1,gset);
    }
    else if (cedge == 2) {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce1,he1));
      es.insert(make_pair(ce3,he3));
      es.insert(make_pair(ce4,he4));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce2.reverse());
      insertGlue(hepos->second,he2,gset);
    }
    else if (cedge == 4) {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce1,he1));
      es.insert(make_pair(ce2,he2));
      es.insert(make_pair(ce4,he4));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce3.reverse());
      insertGlue(hepos->second,he3,gset);
    }
    else {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce1,he1));
      es.insert(make_pair(ce2,he2));
      es.insert(make_pair(ce3,he3));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce4.reverse());
      insertGlue(hepos->second,he4,gset);
    }
  
    // Removes commom edge from the boundary edge set
    es.erase(hepos);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1.reverse(),ledges)) {
      he1->getEdge()->constraintOn();
    }
    if (isConstrainedEdge(ce2.reverse(),ledges)) {
      he2->getEdge()->constraintOn();
    }
    if (isConstrainedEdge(ce3.reverse(),ledges)) {
      he3->getEdge()->constraintOn();
    }
    if (isConstrainedEdge(ce4.reverse(),ledges)) {
      he4->getEdge()->constraintOn();
    }
  }

  // -----------------------------------------------------------------
  // Method makeQuadFaceGlue2()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh and keeps track of its two common edges with other faces
  // of the mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFaceGlue2(
			       const tQuadrilateral2& f,
			       int cv1,
			       int cv2,
			       int cv3,
			       int cv4,
			       const Point2& p1,
			       const Point2& p2,
			       const Point2& p3,
			       const Point2& p4,
			       const tEdgeKey& ce1,
			       const tEdgeKey& ce2,
			       const tEdgeKey& ce3,
			       const tEdgeKey& ce4,
			       int ne,
			       GlueEdgeSet& gset,
			       const ConstrainedEdgeSet& ledges
			       )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));
 
    // Inserts new edge code into boundary edge set and keeps track of
    // the two edges to be glued in the second stage.
    EdgeSet::iterator hepos1, hepos2;
    if (ne == 3) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      es.insert(make_pair(ce3.reverse(),he3));
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
    }
    else if (ne == 5) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce3);
      es.insert(make_pair(ce2.reverse(),he2));
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he3,gset);
    }
    else if (ne == 9) {
      hepos1 = es.find(ce4);
      hepos2 = es.find(ce1);
      es.insert(make_pair(ce2.reverse(),he2));
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he4,gset);
      insertGlue(hepos2->second,he1,gset);
    }
    else if (ne == 6) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce3);
      es.insert(make_pair(ce1.reverse(),he1));
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he3,gset);
    }
    else if (ne == 10) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce4);
      es.insert(make_pair(ce1.reverse(),he1));
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he4,gset);
    }
    else {  // if (ne == 12) {
      hepos1 = es.find(ce3);
      hepos2 = es.find(ce4);
      es.insert(make_pair(ce1.reverse(),he1));
      es.insert(make_pair(ce2.reverse(),he2));
      insertGlue(hepos1->second,he3,gset);
      insertGlue(hepos2->second,he4,gset);
    }
    
    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeQuadFaceGlue3()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh  and keeps  track of its  three common edges  with other
  // faces of the  mesh, so that we can  glue this connected component
  // later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFaceGlue3(
			       const tQuadrilateral2& f,
			       int cv1,
			       int cv2,
			       int cv3,
			       int cv4,
			       const Point2& p1,
			       const Point2& p2,
			       const Point2& p3,
			       const Point2& p4,
			       const tEdgeKey& ce1,
			       const tEdgeKey& ce2,
			       const tEdgeKey& ce3,
			       const tEdgeKey& ce4,
			       int ne,
			       GlueEdgeSet& gset,
			       const ConstrainedEdgeSet& ledges
			       )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    // Inserts new edge code into boundary edge set
    EdgeSet::iterator hepos1, hepos2, hepos3;
    if (ne == 7) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      hepos3 = es.find(ce3);
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
      insertGlue(hepos3->second,he3,gset);
    }
    else if (ne == 11) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      hepos3 = es.find(ce4);
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
      insertGlue(hepos3->second,he4,gset);
    }
    else if (ne == 13) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce3);
      hepos3 = es.find(ce4);
      es.insert(make_pair(ce2.reverse(),he2));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he3,gset);
      insertGlue(hepos3->second,he4,gset);
    }
    else {  // if (ne == 14) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce3);
      hepos3 = es.find(ce4);
      es.insert(make_pair(ce1.reverse(),he1));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he3,gset);
      insertGlue(hepos3->second,he4,gset);
    }

    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);
    es.erase(hepos3);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeQuadFaceGlue4()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh  and keeps  track of its  three common edges  with other
  // faces of the  mesh, so that we can  glue this connected component
  // later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFaceGlue4(
			       const tQuadrilateral2& f,
			       int cv1,
			       int cv2,
			       int cv3,
			       int cv4,
			       const Point2& p1,
			       const Point2& p2,
			       const Point2& p3,
			       const Point2& p4,
			       const tEdgeKey& ce1,
			       const tEdgeKey& ce2,
			       const tEdgeKey& ce3,
			       const tEdgeKey& ce4,
			       GlueEdgeSet& gset,
			       const ConstrainedEdgeSet& ledges
			       )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    // Keeps track of edges to be glued in the second stage
    EdgeSet::iterator hepos1, hepos2, hepos3, hepos4;
    hepos1 = es.find(ce1);
    hepos2 = es.find(ce2);
    hepos3 = es.find(ce3);
    hepos4 = es.find(ce4);

    // Inserts new edge code into boundary edge set and keeps track of
    // the two edges to be glued in the second stage.
    insertGlue(hepos1->second,he1,gset);
    insertGlue(hepos2->second,he2,gset);
    insertGlue(hepos3->second,he3,gset);
    insertGlue(hepos4->second,he4,gset);

    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);
    es.erase(hepos3);
    es.erase(hepos4);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method checkFaceConsistency()
  //
  // Checks  if  a  given  quadrilateral  face  is  geometrically  and
  // topologically consistent.
  //
  // -----------------------------------------------------------------
  bool
  tQuadMesh::checkFaceConsistency(const fcPointer& f) const
  {
    // Checks face topology
    if (f->getHalfEdge()->getFace() != f) return false;

    hePointer he1 = f->getHalfEdge();
    hePointer he2 = he1->getNext();
    hePointer he3 = he2->getNext();
    hePointer he4 = he3->getNext();

    if (he2->getFace() != f) return false;
    if (he3->getFace() != f) return false;
    if (he4->getFace() != f) return false;

    if (he1->getPrev() != he4) return false;
    if (he2->getPrev() != he1) return false;
    if (he3->getPrev() != he2) return false;
    if (he4->getPrev() != he3) return false;

    egPointer e1 = he1->getEdge();
    egPointer e2 = he2->getEdge();
    egPointer e3 = he3->getEdge();
    egPointer e4 = he4->getEdge();

    if (e1->getMate(he1).isNotNull()) {
      hePointer hm = e1->getMate(he1);
      if (hm->getEdge() != e1) return false;
      if (hm->getVertex() != he1->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he1->getVertex()) {
	return false;
      }
    }
    else if (!e1->isConstrained() && !e1->isOnBoundary()) {
      return false;
    }

    if (e2->getMate(he2).isNotNull()) {
      hePointer hm = e2->getMate(he2);
      if (hm->getEdge() != e2) return false;
      if (hm->getVertex() != he2->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he2->getVertex()) {
	return false;
      }
    }
    else if (!e2->isConstrained() && !e2->isOnBoundary()) {
      return false;
    }

    if (e3->getMate(he3).isNotNull()) {
      hePointer hm = e3->getMate(he3);
      if (hm->getEdge() != e3) return false;
      if (hm->getVertex() != he3->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he3->getVertex()) {
	return false;
      }
    }
    else if (!e3->isConstrained() && !e3->isOnBoundary()) {
      return false;
    }
    
    if (e4->getMate(he4).isNotNull()) {
      hePointer hm = e4->getMate(he4);
      if (hm->getEdge() != e4) return false;
      if (hm->getVertex() != he4->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he4->getVertex()) {
	return false;
      }
    }
    else if (!e4->isConstrained() && !e4->isOnBoundary()) {
      return false;
    }

    // Checks face geometry
    Point2 p1 = he1->getVertex()->getPoint();
    Point2 p2 = he2->getVertex()->getPoint();
    Point2 p3 = he3->getVertex()->getPoint();
    Point2 p4 = he4->getVertex()->getPoint();

    tQuadrilateral2 q;

    q.insert(0,1,Point2(p1));
    q.insert(1,2,Point2(p2));
    q.insert(2,3,Point2(p3));
    q.insert(3,4,Point2(p4));

    if (!q.isSimple()) {
      return false;
    }

    return true;
  }

  // -----------------------------------------------------------------
  // Method checkMeshConsistency()
  //
  // Checks   if  this   quadrilateral  mesh   is   geometrically  and
  // topologically consistent.
  //
  // -----------------------------------------------------------------
  bool
  tQuadMesh::checkMeshConsistency() const
  {
    // Checks consistency of each face in the mesh
    for (FIT fi = m_cp->fcBegin(); fi != m_cp->fcEnd(); ++fi) {
      if (!checkFaceConsistency(*fi)) return false;
    }

    return true;
  }

}     // end namespace meshview
